/**
 * Result of `watch(value)` call.
 *
 * @public
 * @typeParam Value - Type of the value.
 * @see {@link watch} for usage examples.
 */
export type WatchResult<Value> = Readonly<{
    /**
     * Proxy object for the input value. Use this to derive data from it while
     * tracking the property access.
     */
    proxy: Value;
    /**
     * The watcher object associated with the `proxy`.
     *
     * @see {@link IWatcher} for API reference.
     */
    watcher: IWatcher;
}>;
/**
 * Result of `watchAll(values)` call.
 *
 * @public
 * @typeParam Value - Type of the value.
 * @see {@link watchAll} for usage examples.
 */
export type WatchAllResult<Values extends ReadonlyArray<unknown>> = Readonly<{
    /**
     * Proxy objects for the input values.
     *
     * @see {@link WatchResult} for details.
     */
    proxies: Values;
    /**
     * The watcher object associated with `proxies`.
     *
     * @see {@link IWatcher} for API reference.
     */
    watcher: IWatcher;
}>;
/**
 * API interface for `watcher` property of `WatchResult` and `WatchAllResult`.
 *
 * @public
 */
export interface IWatcher {
    /**
     * Unwraps the result derived from the input value proxy.
     *
     * The `result` object may contain proxies if sub-objects of the input value
     * proxy were put into the derived data as they are. This method will replace
     * all proxy occurences with their underlying object's value so that the
     * final result is proxy-free.
     *
     * @param result - The derived result object that might contain proxies in its
     *                 sub-properties.
     * @returns The result without proxies in any of its deep sub-properties.
     */
    unwrap<Result>(result: Result): Result;
    /**
     * Stops tracking of property access for the proxies associated with this
     * watcher. Accessing proxy (or sub-proxy) properties after calling
     * `watcher.stop()` will generate a runtime error.
     *
     * Use this to make sure at runtime that no Proxy ends up in your derived
     * data.
     */
    stop(): void;
    /**
     * Returns `true` if tracked sub-properties of `oldValue` are different from
     * (or not present in) the `newValue`. `oldValue` must be either the
     * `value` parameter of `watch` (or one of the `values` of `watchAll`) or
     * a sub-object of it.
     *
     * @see {@link watch}
     * @see {@link watchAll}
     *
     * @param oldValue - `value` argument of `watch`/`watchAll` or its sub-object.
     * @param newValue - any object or primitive to compare against.
     * @returns `true` if tracked properties are different, otherwise - `false`.
     */
    isChanged<Value>(oldValue: Value, newValue: Value): boolean;
}
/**
 * Wraps the `value` into a proxy and returns a `WatchResult` to track the
 * property (and sub-property) access of the object and compare objects.
 *
 * @see {@link WatchResult}
 * @param value - input value which could be a plain object, array, function or
 *                a primitive.
 * @returns WatchResult object that holds the `proxy` for the `value` and
 *          `watcher` object that tracks property access and does the
 *          comparison.
 *
 * @example
 * Here's an example with different objects, but unchanged accessed properties:
 * ```
 * import { watch } from '@indutny/sneequals';
 *
 * const value = { a: { b: 1 } };
 * const { proxy, watcher } = watch(value);
 * const derived = watcher.unwrap({ b: proxy.a.b });
 *
 * // Further access to `proxy` (or its sub-proxies) would throw.
 * watcher.stop();
 *
 * // Prints `{ b: 1 }`
 * console.log(derived);
 *
 * const sameProperties = { a: { b: 1 } };
 *
 * // Prints `false` because these are different objects.
 * console.log(sameProperties === value);
 *
 * // Prints `false` because the tracked `value.a.b` didn't change.
 * console.log(watcher.isChanged(value, sameProperties));
 * ```
 */
export declare const watch: <Value>(value: Value) => Readonly<{
    /**
     * Proxy object for the input value. Use this to derive data from it while
     * tracking the property access.
     */
    proxy: Value;
    /**
     * The watcher object associated with the `proxy`.
     *
     * @see {@link IWatcher} for API reference.
     */
    watcher: IWatcher;
}>;
/**
 * Similar to `watch(value)` this method wraps a list of values with a single
 * `IWatcher` instance and tracks access to properties (sub-properties) of each
 * individual element of the `values` list.
 *
 * @see {@link WatchAllResult}
 * @param values - list of input values that could be plain objects, arrays,
 *                 functions or primitives.
 * @returns WatchAllResult object that holds the `proxies` for the `values` and
 *          `watcher` object that tracks property access and does the
 *          comparison.
 *
 * @example
 * Here's an example with different objects, but unchanged accessed properties:
 * ```
 * import { watchAll } from '@indutny/sneequals';
 *
 * const values = [{ a: { b: 1 } }, { c: 2 }];
 * const { proxies, watcher } = watchAll(value);
 * const derived = watcher.unwrap({ b: proxies[0].a.b, c: proxies[1].c });
 *
 * // Further access to `proxy` (or its sub-proxies) would throw.
 * watcher.stop();
 *
 * // Prints `{ b: 1, c: 2 }`
 * console.log(derived);
 *
 * // Prints `false` because the tracked `value.a.b` didn't change.
 * console.log(watcher.isChanged(values[0], { a: { b: 1 } }));
 *
 * // Prints `true` because the tracked `value.c` changed.
 * console.log(watcher.isChanged(values[1], { c: 3 }));
 * ```
 */
export declare const watchAll: <Values extends readonly unknown[]>(values: Values) => Readonly<{
    /**
     * Proxy objects for the input values.
     *
     * @see {@link WatchResult} for details.
     */
    proxies: Values;
    /**
     * The watcher object associated with `proxies`.
     *
     * @see {@link IWatcher} for API reference.
     */
    watcher: IWatcher;
}>;
/**
 * Options for `memoize()` method.
 *
 * @see {@link memoize} for additional details.
 */
export interface IMemoizeOptions<Params extends ReadonlyArray<unknown>> {
    /**
     * This optional method is called on every cache hit.
     */
    onHit?(): void;
    /**
     * This optional method is called on every cache miss.
     *
     * Note that since `params` were used when creating the `watcher` - you can
     * use them in `watcher` API methods and in `getAffectedPaths`.
     *
     * @param watcher - the newly created `IWatcher` object created with
     *                  `watchAll(params)` API method.
     * @param params - an array of parameters that generated the cache miss.
     * @param pastParams - an array of parameters from the previous cache miss.
     *
     * @see {@link IWatcher}
     * @see {@link watchAll}
     * @see {@link getAffectedPaths}
     */
    onMiss?(watcher: IWatcher, params: Params, pastParams?: Params): void;
}
/**
 * Returns memoized version of the parameter function `fn`. The memoized
 * function will return cached value as long as the arguments of the call have
 * the same tracked values as the last time.
 *
 * Memoized function has two cache levels:
 * - A WeakMap by first object parameter
 * - Global cache that only holds the last execution result.
 *
 * @see {@link IMemoizeOptions} for details on available options.
 *
 * @param fn - function to be memoized
 * @param options - an optional options object
 * @returns memoized function.
 *
 * @example
 * With two parameters:
 * ```
 * import { memoize } from '@indutny/sneequals';
 *
 * const fn = memoize((a, b) => ({ result: a.value + b.value }));
 *
 * // Prints `{ result: 3 }`
 * const answer = fn({ value: 1 }, { value: 2 });
 * console.log(answer);
 *
 * const cachedAnswer = fn({ value: 1 }, { value: 2 });
 *
 * // Even though the input objects are different the `cachedResult` is the
 * // same since the tracked properties didn't change.
 * // Prints `true`.
 * console.log(answer === cachedAnswer);
 * ```
 *
 * @example
 * With react-redux and an id lookup.
 * ({@link https://github.com/dai-shi/proxy-memoize#usage-with-react-redux--reselect | Taken from proxy-memoize}.)
 * ```
 * import { useCallback } from 'react';
 * import { useSelector } from 'react-redux';
 * import { memoize } from '@indutny/sneequals';
 *
 * const Component = ({ id }) => {
 *   const { score, title } = useSelector(useCallback(memoize(state => ({
 *     score: getScore(state),
 *     title: state.titles[id],
 *   })), [id]));
 *   return <div>{score.score} {score.createdAt} {title}</div>;
 * };
 * ```
 */
export declare const memoize: <Params extends readonly unknown[], Result>(fn: (...params: Params) => Result, options?: IMemoizeOptions<Params> | undefined) => (...params: Params) => Result;
/**
 * Returns a list of affected (accessed) JSON paths (and sub-paths) in the
 * `value`. This function is provided mostly as a debugging tool to see which
 * JSON paths might have caused the cache invalidation in `memoize()`.
 *
 * @param watcher - A watcher object that tracked the access to paths/sub-paths
 *                  of the `value`
 * @param value - A `value` tracked by `watcher`
 *
 * @see {@link IWatcher}
 * @see {@link watch} for the details on how to get an `IWatcher` instance.
 * @see {@link memoize}
 *
 * @example
 * ```
 * import { watch, getAffectedPaths } from '@indutny/sneequals';
 *
 * const value = { a: { b: 1 } };
 * const { proxy, watcher } = watch(value);
 *
 * // Prints `1`
 * console.log(proxy.a.b);
 *
 * // Prints `['a.b']`
 * console.log(getAffectedPaths(watcher, value));
 * ```
 */
export declare const getAffectedPaths: (watcher: IWatcher, value: unknown) => ReadonlyArray<string>;
